/*
 * Copyright (C) 1997 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: background.c,v 1.6 1997/12/24 06:41:37 yuo Exp $
 */

#include <math.h>
#include "mgp.h"



/* background gradation */
#define G_PI	3.1415926535897932385
#define G_PI2	1.5707963267948966192

struct g_color *
name2gcolor(colorname)
     u_char *colorname;
{
    Colormap cmap;
    XColor c0, c1;
    struct g_color *color;

    color = (struct g_color *) malloc (sizeof(struct g_color));
    cmap =  DefaultColormap(display, screen);

    if (XLookupColor(display, cmap, colorname, &c1, &c0)) {
	color->r = (c1.red   >> 8) & 0xff;
	color->g = (c1.green >> 8) & 0xff;
	color->b = (c1.blue  >> 8) & 0xff;
    } else {
	fprintf(stderr,"color '%s' unknown. ignored.\n", colorname);
    }	    

    return color;
}

byte *
draw_gradation(cg)
     struct ctrl_grad * cg;
{
    int   bmask[8] = { 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff };
    byte *pic, *pp;
    u_int height, width;
    u_int rot, bits;
    int   i, j, cn;
    int   r, g, b;
    int	gcols;
    struct g_color *ccol, *ncol;
  
    height = cg->ct_height;
    width  = cg->ct_width;
    rot    = cg->ct_direction;
    gcols  = cg->ct_g_colors;
    bits   = 8;

    if ( (pic = (byte *) malloc(width * height * 3 * sizeof(byte)) ) == NULL ){
	fprintf(stderr,"couldn't alloc space for gradation image\n");
	/* return null image, not quit */
	return NULL;
    }

    for (i=0, pp=pic; i < width * height * 3; i++) *pp++ = 0;

    if (rot==0) {   /* un-rotated linear (vertical) gradient */
	for (i=0; i < gcols; i++)
	    cg->colors[i]->y = ((height-1) * i) / (gcols-1);
    
	ccol = cg->colors[0];
	ncol = cg->colors[1];
    
	for (i=0; i < height; i++) {
	    pp = pic + (i * width * 3);

	    /* advance to ncolt pair of colors if we're outside region */
	    cn = 0;
	    while (cg->colors[cn+1]->y < i ){cn++;}
      
	    /* make color cell */
	    ccol = cg->colors[cn];  ncol = cg->colors[cn+1];
	    r = ccol->r +
		((ncol->r - ccol->r) * (i - ccol->y))/(ncol->y - ccol->y);
	    g = ccol->g +
		((ncol->g - ccol->g) * (i - ccol->y))/(ncol->y - ccol->y);
	    b = ccol->b +
		((ncol->b - ccol->b) * (i - ccol->y))/(ncol->y - ccol->y);
	    r = r & bmask[bits-1]; g = g & bmask[bits-1]; b = b & bmask[bits-1];
	    for (j=0; j < width; j++) {
		*pp++ = (byte) r;  *pp++ = (byte) g;  *pp++ = (byte) b;
	    }
	}
    } else g_rotate(pic, cg);

    return pic;
}

static double tant1, cost1;
static double cost, sint;
static double cos2t, sin2t;
static double dcost, dsint;

/* rotate graphic */
void g_rotate(pic, cg)
     byte *pic;
     struct ctrl_grad * cg;
{
    byte *pp;
    double maxd, mind, del, d, rat, crat, cval, tval;
    double theta, dy, ey, td1, td2;
    int    x, y, cx, cy, r, g, b, bc, nc1;
    int    width, height, rot, mode;
    struct g_color * c1;
    struct g_color * c2;

    width = cg->ct_width;
    height= cg->ct_height;
    rot   = cg->ct_direction;
    mode  = cg->ct_mode;
  
    cx = width/2;  cy = height/2;
    theta = (double) rot * G_PI / 180.0;
#if 0
    tval =  (double) G_PI2 - theta;
    tant1 = tan(tval);
#endif
    cost  = cos(theta);
    sint  = sin(theta);
    dsint = sint*sint;
    dcost = cost*cost;
    sin2t = sin(2*theta);
    cos2t = cos(2*theta);

    nc1 = cg->ct_g_colors - 1;

    /* compute max/min distances */
    if (rot > 0 && rot < 90) {
	mind = cdist(0, 0, cx, cy, rot, mode);
	maxd = cdist(width-1, height-1, cx, cy, rot, mode);
    }
    else if (rot >= 90 && rot < 180) {
	mind = cdist(0, height-1, cx, cy, rot, mode);
	maxd = cdist(width-1, 0,  cx, cy, rot, mode);
    }
    else if (rot >= 180 && rot < 270) {
	mind = cdist(width-1, height-1, cx, cy, rot, mode);
	maxd = cdist(0, 0, cx, cy, rot, mode);
    }
    else {
	mind = cdist(width-1,  0, cx, cy, rot, mode);
	maxd = cdist(0, height-1, cx, cy, rot, mode);
    }
  
    del = maxd - mind;         /* maximum distance */
  
    /* loop start */

    td1 = cx - cx*dcost;
    td2 = cy + 0.5*cx*sin2t;

    for (y=0; y < height; y++) {
	pp = pic + (y * width * 3);
	dy = (cy-y)*0.5*sin2t + td1;
	ey = (y-cy)*dsint + td2;

	for (x=0; x < width; x++) {
	    d = lcdist(x, y, cx, cy, rot, mode, dy, ey);
	    rat = (d - mind) / del;

	    if (rat < 0.0){
		cval = bc = crat = 0.0;
	    } else if (rat > 1.0){
		cval = bc = (double)nc1;
		crat = 0.0;
	    } else {
		cval = rat * nc1;
		bc   = floor(cval);
		crat = cval - bc;
	    }

	    if (bc < nc1) {
		c1 = (cg->colors)[bc];
		c2 = (cg->colors)[bc+1];

		r = c1->r + crat * ((c2->r) - (c1->r));
		g = c1->g + crat * ((c2->g) - (c1->g));
		b = c1->b + crat * ((c2->b) - (c1->b));
	    } else {
		c1 = (cg->colors)[nc1];
		r = c1->r; g = c1->g; b = c1->b;
	    }
	    *pp++ = (byte) r;  *pp++ = (byte) g;  *pp++ = (byte) b;
	}
    }
}

/* compute distance */
double cdist(x, y, cx, cy, rot, mode)
     int x, y, cx, cy, rot, mode;
{
    double x1, y1, x2, y2, x3, d ;
  
    /* special case */
    if (rot == 0)   return (double) (y - cy);
    if (rot == 90)  return (double) (x - cx);
    if (rot == 180) return (double) (cy - y);
    if (rot == 270) return (double) (cx - x);

/* experimental routine for circle patern gradation */
    if (mode == 1){
	d = sqrt((y-cy)*(y-cy)*cost + (x-cx)*(x-cx)*sint);
	if (x + y - cy < 0) d = -d;
	return d;
    };

    /* x1,y1 = original point */
    x1 = (double) x;  y1 = (double) y;
    x2 = cx + (cy-y1)*0.5*sin2t - (cx-x1)*dcost;
    y2 = cy - (cy-y1)*dsint + (cx-x1)*0.5*sin2t;
    x3 = x1-x2;

    d = sqrt(x3*x3 + (y1 - y2)*(y1 - y2));
    if ((rot < 180 && x3<0) || (rot > 180 && x3>0)) d = -d;
    return d;
}

double lcdist(x, y, cx, cy, rot, mode, dy, ey)
    int x, y, cx, cy, rot, mode;
    double dy,ey;
{
    double x1, y1, x2, y2, x3, d ;
  
    /* special case */
    if (rot == 0)   return (double) (y - cy);
    if (rot == 90)  return (double) (x - cx);
    if (rot == 180) return (double) (cy - y);
    if (rot == 270) return (double) (cx - x);

/* experimental routine for circle patern gradation */
    if (mode == 1){
	d = sqrt((y-cy)*(y-cy)*cost + (x-cx)*(x-cx)*sint);
	if (x + y - cy < 0) d = -d;
	return d;
    };

    x1 = (double) x;
    y1 = (double) y;

    x2 = dy + x1*dcost;
    y2 = ey - x1*0.5*sin2t;
    x3 = x1-x2;

    d = sqrt(x3*x3 + (y1 - y2)*(y1 - y2));
    if ((rot < 180 && x3<0) || (rot > 180 && x3>0)) d = -d;
    return d;
}


Image *
make_XImage(pic, width, height)
     byte *pic;
     unsigned int width, height;
{
    Image        *image;
    int           size;

#if USE_COLOR_DEPTH
    byte         *destptr;
    unsigned int  y, maxval;
#endif

    size= height * width * 3;

    /* make Image structure (stolen from image/new.c)*/
    image= (Image *)lmalloc(sizeof(Image));
    image->type= ITRUE;
    image->title= NULL;
    image->rgb.used= image->rgb.size= 0;
    image->width= width;
    image->height= height;
    image->depth= 24;
    image->pixlen= 3;
    image->data= pic;

#if USE_COLOR_DEPTH
#define PM_SCALE(a, b, c) (long)((a) * (c))/(b)
    /* color depth (future use) */
    maxval=0xff;

    destptr= image->data;
    for (y= 0; y < size; y++) {
	*destptr= PM_SCALE(*destptr, maxval, 0xff);
	destptr++;
    }
#endif

    return(image);
}
