/*
 * Copyright (C) 1997 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: font.c,v 1.10 1997/12/24 03:50:01 onoe Exp $
 */

#include "mgp.h"

#ifdef VFLIB

int vfcachesize = 3000;	/*XXX*/
#define	VFHASH_SIZE	255

static char *vf_curname;
static int vfont_fd = -1;
static int vfcinitdone;
static int vfcachecnt;
static struct vfont vfclru;
static struct vfont vfcache[VFHASH_SIZE];
#define	VFCACHE_HASH(w, h, c)	(((c ^ w) ^ h) % VFHASH_SIZE)

int vfcachehit = 0;
int vfcachemiss = 0;

static void initrevbitmap __P((void));

/*
 * revbitmap: reverse the MSB and LSB of a byte value.
 *	example: revbitmap[0x80] == 0x01, revbitmap[0x55] == 0xaa
 * leftbitmap: the index of leftmost bit in a byte value. (MSB is on the left)
 *	example: leftbitmap[0x80] == 0, leftbitmap[0x19] == 3
 * rightbitmap: the index of rightmost bit in a byte value.
 *	example: rightbitmap[0x80] == 0, rightbitmap[0x19] == 7
 */
static u_char *revbitmap = NULL;
static char *leftbitmap = NULL;
static char *rightbitmap = NULL;

static void
initrevbitmap()
{
	int i;
	int j;
	u_char x;
	int right;
	int left;

	revbitmap = (u_char *)malloc(256 * sizeof(u_char));
	leftbitmap= (char *)malloc(256 * sizeof(char));
	rightbitmap= (char *)malloc(256 * sizeof(char));
	if (!revbitmap || !leftbitmap || !rightbitmap) {
		perror("malloc");
		exit(-1);
	}
	for (i = 0; i < 256; i++) {
		right = left = -1;
		x = 0x00;
		for (j = 0; j < 8; j++) {
			if (i & (0x80 >> j)) {
				x |= (1 << j);
				right = j;
				if (left == -1)
					left = j;
			}
		}
		revbitmap[i] = x;
		rightbitmap[i] = right;
		leftbitmap[i] = left;
	}
}

#define	VFC_INSHASH(vfc) {						\
	struct vfont *h, *n;						\
	h = &vfcache[VFCACHE_HASH(vfc->code, vfc->width, vfc->height)];	\
	n = h->next; vfc->next = n; n->prev = vfc;			\
	h->next = vfc; vfc->prev = h;					\
}
#define	VFC_DELHASH(vfc) {						\
	struct vfont *n, *p;						\
	n = vfc->next; p = vfc->prev;					\
	n->prev = p; p->next = n;					\
}
#define	VFC_INSLRU(vfc) {						\
	struct vfont *p;						\
	p = vfclru.lruprev; vfc->lruprev = p; p->lrunext = vfc;		\
	vfclru.lruprev = vfc; vfc->lrunext = &vfclru;			\
}
#define	VFC_DELLRU(vfc) {						\
	struct vfont *n, *p;						\
	n = vfc->lrunext; p = vfc->lruprev;				\
	n->lruprev = p; p->lrunext = n;					\
}

static void
vfc_init()
{
	int i;
	struct vfont *vfc;

	vfc = vfcache;
	for (vfc = vfcache, i = 0; i < VFHASH_SIZE; vfc++, i++)
		vfc->next = vfc->prev = vfc;
	vfclru.lrunext = vfclru.lruprev = &vfclru;
	vfcinitdone++;
}

static void
vfc_free(vfc)
	struct vfont *vfc;
{
	struct vfont *p, *n;

	VFC_DELHASH(vfc);
	VFC_DELLRU(vfc);
	XDestroyImage(vfc->image);
#ifdef ANTIALIAS
	free(vfc->dbitmap);
#endif
	free(vfc);
	vfcachecnt--;
}

static struct vfont *
vfc_lookup(code, width, height, fontname)
	u_int code, width, height;
	char *fontname;
{
	struct vfont *vfc, *hvfc;
	struct vfont *p, *n;

	hvfc = &vfcache[VFCACHE_HASH(code, width, height)];
	for (vfc = hvfc->next; vfc != hvfc; vfc = vfc->next) {
		if (vfc->code == code &&
		    vfc->width == width &&
		    vfc->height == height &&
		    vfc->fontname == vf_curname) {
			vfcachehit++;
			VFC_DELLRU(vfc);
			VFC_INSLRU(vfc);
		}
	}
	vfcachemiss++;
	return NULL;
}

static struct vfont *
vfc_alloc(code, width, height, fontname)
	u_int code, width, height;
	char *fontname;
{
	u_long *vfdata;
	u_int rast = (width + 7) / 8;
#ifdef ANTIALIAS
	u_int rast2 = (width * 2 + 7) / 8;
	u_char *lbitmap;
	u_char *dbitmap;
#endif
	u_int char_len;
	struct vfont *vfc;
	u_int max_x, min_x;
	u_int ix, iy;
	u_char data, *bitmap;
	u_char *s, *d;
	u_int bsize;
	XImage *xim;

	bsize = rast * height;
	vfc = (struct vfont *)malloc(sizeof(*vfc));
	if (vfc == NULL) {
		fprintf(stderr, "vfc_alloc: malloc failed\n");
		exit(1);
	}
	vfc->code = code;
	vfc->width = width;
	vfc->height = height;
	vfc->fontname = vf_curname;
	bitmap = (u_char *)malloc(bsize);
	if (bitmap == NULL) {
		fprintf(stderr, "vfc_alloc: malloc failed\n");
		exit(-1);
	}
#ifdef ANTIALIAS
/*XXX: TODO: cleanup, optimization */
	lbitmap = (u_char *)malloc(rast2 * height * 2);
	dbitmap = (u_char *)malloc(width * height);
	if (lbitmap == NULL || dbitmap == NULL) {
		fprintf(stderr, "vfc_alloc: malloc failed\n");
		exit(-1);
	}
	memset(lbitmap, 0, rast2 * height * 2);
	VF_GetBitmap(code, vfont_fd, width * 2, height * 2, rast2, 0, lbitmap);

	{
	int i, j, x1, x2, y, b1, b2;
	u_char *s1, *s2, *d, *z;
	static u_char dp[] = { 0, 1, 1, 2 };

	z = dbitmap;
	d = bitmap;
	for (j = 0; j < height; j++) {
		s1 = lbitmap + rast2 * (j * 2);
		s2 = s1 + rast2;
		y = 0;
		for (i = 0, b1 = 7, b2 = 6; i < width; i++, b1--, b2 -= 2) {
			if (b1 < 0) {
				*d++ = y;
				y = 0;
				b1 = 7;
			}
			if (b2 < 0) {
				s1++, s2++;
				b2 = 6;
			}
			x1 = (*s1 >> b2) & 0x03;
			x2 = (*s2 >> b2) & 0x03;
			if (x1 | x2)
				y |= 1 << b1;
			*z++ = dp[x1] + dp[x2];
		}
		*d++ = y;
		for (i += 2; i < rast * 2; i++)
			*d++ = 0;
	}
	}
	free(lbitmap);
	vfc->dbitmap = dbitmap;
#else /* ANTIALAIS */
	memset(bitmap, 0, bsize);
	if (!(mgp_flag & FL_OUTLINE)) {
		VF_GetBitmap(code, vfont_fd, width, height, rast, 0, bitmap);
	} else {  /* OR */
		if ((vfdata = VF_GetOutline(code, vfont_fd)) == NULL) {
			fprintf(stderr, "Failed to get outline : %04x\n", code);
		} else {
			VF_DrawOutline(vfdata, vfont_fd, width, height, rast,
				0, bitmap);
			VF_FreeOutline(vfdata, vfont_fd);
		}
	}
#endif /* ANTIALAIS */

	if (!revbitmap)
		initrevbitmap();

	/* compute right/left boundary */
	max_x = -1;
	min_x = -1;
	for (ix = 0; ix < rast; ix++) {
		data = 0x00;
		s = bitmap + ix;
		for (iy = 0, s = bitmap + ix; iy < height; iy++, s += rast)
			data |= *s;
		if (data) {
			max_x = ix * 8 + rightbitmap[data];
			if (min_x == -1)
				min_x = ix * 8 + leftbitmap[data];
		}
	}
	/* XXX: ".," ... */
	if (code >= 0x2121 && code <= 0x2125 && max_x < width / 2)
		max_x += width / 2;

	xim = XCreateImage(display, DefaultVisual(display, screen), 1,
		XYBitmap, 0, NULL, width, height, 8, 0);
	if (xim == NULL) {
		fprintf(stderr, "vfc_alloc: XCreateImage failed\n");
		exit(1);
	}
	if (xim->bitmap_bit_order == LSBFirst) {
		/* convert the bit order of image bitmap */
		for (iy = 0, s = bitmap; iy < height; iy++) {
			for (ix = 0; ix < rast; ix++, s++) {
				*s = revbitmap[*s];
			}
		}
	}
	xim->data = bitmap;
	vfc->image = xim;
	/* horizontal gap of 1 dot is included by default */
	char_len = max_x - min_x + 2;
	vfc->minx = min_x;
	vfc->charlen = char_len;

	VFC_INSHASH(vfc);
	VFC_INSLRU(vfc);
	vfcachecnt++;
	return vfc;
}

struct vfont *
vfc_get(code, width, height, force)
	u_int code, width, height;
	int force;
{
	struct vfont *vfc;

	if (!vfcinitdone)
		vfc_init();
	vfc = vfc_lookup(code, width, height, vf_curname);
	if (vfc == NULL) {
		if (vfcachecnt >= vfcachesize) {
			if (!force)
				return NULL;
			while (vfcachecnt >= vfcachesize)
				vfc_free(vfclru.lrunext);
		}
		vfc = vfc_alloc(code, width, height, vf_curname);
	}
#if 0
printf("vfc_get: cache=%d, hit=%d, miss=%d\n", vfcachecnt, vfcachehit, vfcachemiss);
#endif
	return vfc;
}

void
vfc_setfont(fontname)
	char *fontname;
{
	if (vfont_fd < 0) {
		if (VF_Init(vfcap_name) < 0) {
			fprintf(stderr, "Error VF_Init\n");
			exit(-1);
		}
	} else {
		VF_CloseFont(vfont_fd);
		VF_Deinit();
	}
	vfont_fd = VF_OpenFont(fontname);
	vf_curname = fontname;
	if (vfont_fd < 0) {
		fprintf(stderr, "Can't open vffont %s\n", fontname);
		exit(-1);
	}
}

#ifdef ANTIALIAS
XImage *
vfc_image(vfc, fore, back)
	struct vfont *vfc;
	long fore, back;
{
	int verbose = mgp_flag & FL_VERBOSE;
	int x, y;
	static XColor col[5];
	Colormap cmap;
	XImage *xim;
	Visual *visual;
	u_char *s;

	visual = DefaultVisual(display, screen);
	if (fore != col[4].pixel || back != col[0].pixel) {
		col[4].pixel = fore;
		col[0].pixel = back;

		cmap = DefaultColormap(display, screen);
		col[0].flags = col[4].flags = DoRed|DoGreen|DoBlue;
		col[1].flags = col[2].flags = col[3].flags = 0;
		XQueryColors(display, cmap, col, 5);
		for (x = 3; x > 0; x--) {
			col[x].red   = (col[4].red  *x + col[0].red  *(4-x)) /4;
			col[x].green = (col[4].green*x + col[0].green*(4-x)) /4;
			col[x].blue  = (col[4].blue *x + col[0].blue *(4-x)) /4;
			if (!XAllocColor(display, cmap, &col[x])) {
				if (verbose)
					printf("vfc_image: cannot allocate color for level %d (using %d)\n", x, x + 1);
				col[x].pixel = col[x + 1].pixel;
			}
		}
	}

	xim = XCreateImage(display, visual, depth, ZPixmap, 0, NULL, 
			vfc->width, vfc->height, depth == 24 ? 32 : depth, 0);
	if (xim == NULL)
		return NULL;
	xim->data = malloc(xim->bytes_per_line * xim->height);
	if (xim->data == NULL) {
		XDestroyImage(xim);
		return NULL;
	}

	/* XXX: need optimization */
	s = vfc->dbitmap;
	for (y = 0; y < vfc->height; y++) {
		for (x = 0; x < vfc->width; x++) {
			XPutPixel(xim, x, y, col[*s++].pixel);
		}
	}
	return xim;
}
#endif /* ANTIALIAS */

#endif /* VFLIB */
