/*
 * Copyright (C) 1997 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: mgp.c,v 1.51 1997/12/24 06:41:40 yuo Exp $
 */

static char *mgp_version = "1.00a (971224)";

#include "mgp.h"

Window plwin[MAXPAGE];
Pixmap maskpix;
XFontStruct *plfs;
XFontStruct *plkfs;

u_int pause_flag;

u_int pg_mode;
u_int pl_mode;
u_int pl_tmode;
u_int pl_nx, pl_ny, pl_y;
u_long pl_fh, pl_fw;

int rakugaki = 0;
int rakugaki_x = -1;
int rakugaki_y = -1;

u_long depth_mask;

u_char back_clname[MAXPATHLEN];

#ifdef VFLIB
char *vfcap_name;
#endif

static void mgp_usage __P((char *));
static void mgp_show_version __P((char *));
static void beep __P((void));
static void main_loop __P((u_int));
static void set_geometry __P((u_char *));
static void rakugaki_update __P((XEvent *));

/*image*/
extern char *expandPath __P((char *));

int
main(argc, argv)
	int argc;
	char **argv;
{
	u_int opt;
	extern char *optarg;
	extern int optind;
	u_char *progname;
	u_int start_page = 1;
	char buf[BUFSIZ], *p;
	char *vfont_name;
	extern char *Paths[];
	extern int NumPaths;

	progname = argv[0];

	while ((opt = getopt(argc, argv, "ABdvVob:c:g:f:hlGp:qt:Q:")) != -1) {
		switch (opt) {
		case 'B':
			mgp_flag |= FL_BIMAGE;
			break;

		case 'd':
			mgp_flag |= FL_DEMO;
			break;

		case 'V':
			mgp_flag |= FL_VERBOSE;
			break;

		case 'l':
			mgp_flag |= FL_OUTLINE;
			break;

		case 'G':
			pg_mode = 1;
			break;

		case 'o':
			mgp_flag |= FL_OVER;
			break;

#ifdef VFLIB
		case 'c':
			vfcap_name = optarg;
			break;

		case 'f':
			vfont_name = optarg;
			break;
#ifdef ANTIALIAS
		case 'A':
			mgp_flag |= FL_NOALIAS;
			break;
#endif
#endif /*VFLIB*/

		case 'b':
			strcpy(back_clname, optarg);
			break;

		case 'p':
			start_page = atoi(optarg);
			break;

		case 'g':
			set_geometry(optarg);
			break;

		case 'q':
			mgp_flag |= FL_NOBEEP;
			break;

		case 't':
			t_fin = atoi(optarg);
			tbar_flag = 1;
			break;

		case 'Q':
			b_quality = atoi(optarg);
			quality_flag = 1;
			break;

		case 'v':
			mgp_show_version(progname);
			/*NOTREACHED*/

		case 'h':
		default:
			mgp_usage(progname);
			/*NOTREACHED*/
		}
	}

	argc -= optind;
	argv += optind;

	if (argc != 1) {
		mgp_usage(progname);
		/*NOTREACHED*/
	}
	mgp_fname = argv[0];

	init_win1();
	strncpy(buf, mgp_fname, sizeof(buf));
	if (p = rindex(buf, '/')) {
		*p = '\0';
        	Paths[NumPaths++]= expandPath(buf);
	}
	loadPathsAndExts();
	load_file(mgp_fname);
	if (parse_error)
		exit(-1);
	init_win2();
#ifdef VFLIB
	if (vfcap_name == NULL)
		vfcap_name = VFCAP;
	if (vfont_name == NULL)
		vfont_name = VFONT;
	vfc_setfont(vfont_name);
#endif

	if (mgp_flag & FL_DEMO) {
		while (start_page <= maxpage)
			draw_page(start_page++);
	} else
		main_loop(start_page);

	exit(0);
}

static void
mgp_show_version(name)
	char *name;
{
	char *p;

	if (p = strrchr(name, '/'))
		p ++;
	else
		p = name;
	fprintf(stderr, "%s version %s\n", p, mgp_version);
	exit(0);
}
	
static void
mgp_usage(name)
	char *name;
{
	char *p;

	if (p = strrchr(name, '/'))
		p ++;
	else
		p = name;
	fprintf(stderr, "Usage: %s [opts] mgpfile\n", p);
	fprintf(stderr, "\t-B: Ignore background image\n");
	fprintf(stderr, "\t-d: Demo page is prepended\n");
	fprintf(stderr, "\t-V: Verbose\n");
	fprintf(stderr, "\t-l: Do not use outline font\n");
	fprintf(stderr, "\t-o: Controlled by the window manager\n");
	fprintf(stderr, "\t-G: Page guide is on\n");
#ifdef VFLIB
	fprintf(stderr, "\t-c <vfcap>: Specify vfontcap file\n");
	fprintf(stderr, "\t-f <font>: Specify fontname\n");
#endif /*VFLIB*/
	fprintf(stderr, "\t-b <color>: Specify background color\n");
	fprintf(stderr, "\t-p <page>: Start at the specified page\n");
	fprintf(stderr, "\t-g <geometry>: Set window geometry\n");
	fprintf(stderr, "\t-Q <quality>: Set background image quality(0-100)\n");
	fprintf(stderr, "\t-q Do not beep on errors\n");
	fprintf(stderr, "\t-h Show this usage\n");
	fprintf(stderr, "\t-v Show version\n");	
	exit(0);
}

static void
beep()
{
	if (!(mgp_flag & FL_NOBEEP))
		XBell(display, 0);
}

static void
main_loop(start_page)
	u_int start_page;
{
	XEvent e;
	KeySym key;
	u_int page;
	u_int i;
	u_int number = 0;
	u_int shift = 0;
	u_int control = 0;
	static Cursor pen_curs;

	if (!pen_curs)
		pen_curs = XCreateFontCursor(display, XC_pencil);
	page = start_page;
	draw_page(page);

	while (1) {
		if (rakugaki)
			XDefineCursor(display, window, pen_curs);
		else
			XUndefineCursor(display, window);
		XNextEvent(display, &e);
		switch (e.type) {
		case EnterNotify:
			for (i = 1; i <= maxpage; i++) {
				if (e.xany.window == plwin[i]) {
					XClearWindow(display, plwin[i]);
					pl_pdraw(i, gc_plrev);
					pl_title(i);
				}
			}
			XFlush(display);
			break;

		case LeaveNotify:
			for (i = 1; i <= maxpage; i++) {
				if (e.xany.window == plwin[i]) {
					XClearWindow(display, plwin[i]);
					pl_pdraw(i, gc_pl);
					pl_title(0);
				}
			}
			XFlush(display);
			break;

		case ButtonPress:
			if (rakugaki && e.xany.window == window) {
				rakugaki_update(&e);
				break;
			}

			for (i = 1; i <= maxpage; i++) {
				if (e.xany.window == plwin[i]) {
					page = i;
					pause_flag = 0;
					pl_off();
					draw_page(page);
					pl_tmode = 0;
					break;
				}
			}
			if (e.xany.window == window) {
				if (t_start == 0) t_start = time(0);
				if (e.xbutton.button == 1) {
					if (shift)
						pause_flag = 0;
					if (!pause_flag) {
						if (page < maxpage) {
							page ++;
							draw_page(page);
						} else
							beep();
					} else
						draw_page(page);
				} else {
					if (e.xbutton.button == 3) {
						pause_flag = 0;
						if (page > 1) {
							page --;
							draw_page(page);
						} else
							beep();
					}
				}
			}
			break;

		case ButtonRelease:
			if (rakugaki && e.xany.window == window)
				rakugaki_update(&e);
			break;

		case MotionNotify:
			if (rakugaki && e.xany.window == window)
				rakugaki_update(&e);
			break;

		case KeyPress:
			key = XLookupKeysym((XKeyEvent *)&e, 0);

			switch (key) {
			case XK_q:
			case XK_Escape:
				exit(0);
				break;

			case XK_x:
			    {
				rakugaki = 1 - rakugaki;
				rakugaki_x = rakugaki_y = -1;

				XSelectInput(display, window,
					rakugaki ? EVENT_RAKUGAKI
						 : EVENT_DEFAULT);
				if (rakugaki) {
					XDefineCursor(display, window,
						pen_curs);
				} else
					XUndefineCursor(display, window);
				XFlush(display);
				break;
			    }

			case XK_f:
			case XK_j:
			case XK_n:
			case XK_Down:
			case XK_space:
				if (t_start == 0) t_start = time(0);
				if (number == 0) number = 1;
				if (!pause_flag) {
					if (page + number <= maxpage) {
						page += number;
						draw_page(page);
					} else
						beep();
				} else
					draw_page(page);
				number = 0;
				break;

			case XK_b:
			case XK_k:
			case XK_p:
			case XK_Up:
			case XK_BackSpace:
			case XK_Delete:
				pause_flag = 0;
				if (number == 0) number = 1;
				if (page - number >= 1) {
					page -= number;
					draw_page(page);
				} else
					beep();
				number = 0;
				break;

			case XK_l:
				if (!control)
					break;
				draw_page(page);
				number = 0;
				break;

			case XK_Control_L:
			case XK_Control_R:
				pl_on();
				control = 1;
				number = 0;
				break;

			case XK_Shift_L:
			case XK_Shift_R:
				shift = 1;
				number = 0;
				break;
#ifdef VFLIB
#ifdef ANTIALIAS
			case XK_a:
				if (shift) {
					mgp_flag ^= FL_NOALIAS;
					draw_page(page);
				}
				number = 0;
				break;
#endif
#endif

			case XK_g:
				if (shift) {
					pg_mode = 1 - pg_mode;
					if (pg_mode)
						pg_on(page, 1);
					else
						pg_off();
				} else {
					if (number <= maxpage && number > 0) {
						page = number;
						draw_page(page);
					} else
						beep();
				}
				number = 0;
				break;

			case XK_t:	/* toggle timebar switch */
				if (tbar_flag)
					tbar_flag = 0;
				else if (t_fin)
					tbar_flag = 1;
				break;

			case XK_0:
			case XK_1:
			case XK_2:
			case XK_3:
			case XK_4:
			case XK_5:
			case XK_6:
			case XK_7:
			case XK_8:
			case XK_9:
				number = number * 10 + key - XK_0;
				break;
			default:
				number = 0;
				break;
			}

			break;

		case KeyRelease:
			key = XLookupKeysym((XKeyEvent *)&e, 0);
			switch (key) {
			case XK_Control_L:
			case XK_Control_R:
				pl_off();
				control = 0;
				break;
			case XK_Shift_L:
			case XK_Shift_R:
				shift = 0;
				break;
			}
			break;

		case ConfigureNotify:
			pl_config();
			pause_flag = 0;
			window_width = e.xconfigure.width;
			window_height = e.xconfigure.height;
			draw_page(page);
			break;
		}
	}
}

static void
set_geometry(buf)
	u_char *buf;
{
	u_char *p, *p2, tmp[10];

	p = buf;
	strncpy(tmp, p, strlen(p) - strlen(strchr(p, 'x')));
	window_width = atoi(tmp);

	p = strchr(p, 'x') +1;
	p2 = tmp;
	while (*p != '\0' && *p != '+' && *p != '-')
		*(p2++) = *(p++);
	*p2 = 0;
	window_height = atoi(tmp);
	if (*p == '\0')
		return;

	p2 = tmp;
	*(p2++) = *(p++);
	while (*p != '\0' && *p != '+' && *p != '-')
		*(p2++) = *(p++);
	*p2 = 0;

	window_x = atoi(tmp);
	window_y = atoi(p);
}

/* XXX should think about color (= gc) to work with */
static void
rakugaki_update(e)
	XEvent *e;
{
	int x, y;

	if (e->type == MotionNotify) {
		XMotionEvent *em;
		em = (XMotionEvent *)e;
		x = em->x; y = em->y;
	} else if (e->type == ButtonPress) {
		XButtonPressedEvent *eb;
		eb = (XButtonPressedEvent *)e;
		x = eb->x; y = eb->y;
	} else {
		rakugaki_x = rakugaki_y = -1;
		return;
	}

	if (rakugaki_x < 0 || rakugaki_y < 0)
		XDrawRectangle(display, window, gcpen, x, y, 2, 2);
	else {
		XDrawLine(display, window, gcpen,
			x, y,
			rakugaki_x, rakugaki_y);
		XDrawLine(display, window, gcpen,
			x + 1, y,
			rakugaki_x + 1, rakugaki_y);
		XDrawLine(display, window, gcpen,
			x, y + 1,
			rakugaki_x, rakugaki_y + 1);
		XDrawLine(display, window, gcpen,
			x + 1, y + 1,
			rakugaki_x + 1, rakugaki_y + 1);
	}
	rakugaki_x = x;
	rakugaki_y = y;
}
