/*
 * Copyright (C) 1997 WIDE Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by WIDE Project and
 *	its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/*
 * $Id: mgp.h,v 1.55 1997/12/24 06:41:41 yuo Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <string.h>
#include <sys/param.h>
#include <ctype.h>
#include <assert.h>
#include <time.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <X11/Xatom.h>
#include "image/xloadimage.h"

#ifndef __P
#define	__P(X)	()
#endif

#ifndef	VFLIB
#define USE_XDRAWSTRING
#else
#define USE_XDRAWSTRING_ONLY_SMALL
#define	ANTIALIAS
#endif

#ifdef VFLIB
#include "VF.h"
#define VFONT	"minsl"

#ifndef VFCAP
#define VFCAP	"/usr/local/lib/VFlib/vfontcap"
#endif

#endif /* end of VFLIB */

#define DEFAULT_FORE	"yellow"
#define DEFAULT_BACK	"black"
#define PAGELIST_FONT	"a14"
#define PAGELIST_KFONT	"k14"

#define FONT_FORMAT	"-adobe-%s-*-%d-*-*-*-*-*-iso8859-1"
#define	DEFAULT_X_FONT	"times-medium-r"
#define	CUTIN_DELAY	15000
#define	SHRINK_DELAY	00500

#define	XRESOURCE_FILE	"/usr/X11R6/lib/X11/app-default/Mgp"

#define MAXPAGE		200
#define MAXLINE		100
#define MAXVALLEN	512
#define MAXDIREC	16
#define MAXARG		16
#define MAXTAB		5

#define SP_NONE		0
#define SP_SHRINK	1
#define SP_LCUTIN	2
#define SP_RCUTIN	3

#define	AL_CENTER	1
#define	AL_LEFT		2
#define	AL_RIGHT	3

#define DEFAULT_GRADSTART "blue"
#define DEFAULT_GRADEND "black"
#define DEFAULT_GRADDEPTH 8
#define DEFAULT_GRADCOLORS 128

/* 	mgp command line flags 	*/
#define	FL_OVER		0x01
#define	FL_BIMAGE	0x02
#define	FL_DEMO		0x04
#define	FL_VERBOSE	0x08
#define	FL_OUTLINE	0x10
#define	FL_NOBEEP	0x20
#define	FL_NOALIAS	0x40

/* 	page attribute flags 	*/
#define PGFLAG_NODEF	0x01	/* nodefault */

#define EVENT_DEFAULT \
	(KeyPressMask|KeyReleaseMask|ButtonPressMask|StructureNotifyMask)
#define EVENT_RAKUGAKI \
	(EVENT_DEFAULT|ButtonReleaseMask|Button1MotionMask)

#include "ctlwords.h"

struct ctrl_int {
	u_int ct_value;
};

struct ctrl_long {
	u_long ct_value;
};

struct ctrl_char {
	u_char *ct_value;
};

struct ctrl_image {
	u_char *ct_fname;
	u_int ct_numcolor;
	u_int ct_ximagesize;
	u_int ct_yimagesize;
	u_int ct_zoomflag;
};

/* for gradation image generation*/
#define G_MAXCOLS  24

extern u_int	b_quality;
extern u_int	quality_flag;

struct  g_color { 
  int   r,g,b; 
  int   y; 
};

struct ctrl_grad {
	u_int ct_numcolor;
	int ct_direction;
	u_int ct_width;
	u_int ct_height;
	u_int ct_zoomflag;
	u_int ct_xzoom;
	u_int ct_yzoom;
        u_int ct_depth;
        int   ct_g_colors;
        int   ct_mode;
        struct g_color * colors[G_MAXCOLS];
};

struct ctrl_bar {
	u_long ct_color;
	u_int ct_width;
	u_int ct_start;
	u_int ct_length;
};

struct ctrl {
	u_char ct_op;
	u_char ct_flag;
	u_int ct_page;
	struct ctrl *ct_next;
	union {
		struct ctrl_int ctrl_int;
		struct ctrl_long ctrl_long;
		struct ctrl_char ctrl_char;
		struct ctrl_image ctrl_image;
		struct ctrl_grad ctrl_grad;
		struct ctrl_bar ctrl_bar;
	} ct_val;
};

#define	cti_value	ct_val.ctrl_int.ct_value
#define	ctl_value	ct_val.ctrl_long.ct_value
#define	ctc_value	ct_val.ctrl_char.ct_value

#define	ctm_fname	ct_val.ctrl_image.ct_fname
#define	ctm_numcolor	ct_val.ctrl_image.ct_numcolor
#define	ctm_ximagesize	ct_val.ctrl_image.ct_ximagesize
#define	ctm_yimagesize	ct_val.ctrl_image.ct_yimagesize
#define	ctm_zoomflag	ct_val.ctrl_image.ct_zoomflag

#define ctd_colors	ct_val.ctrl_grad.colors
#define ctd_g_colors	ct_val.ctrl_grad.ct_g_colors
#define	ctd_numcolor	ct_val.ctrl_grad.ct_numcolor
#define ctd_dir		ct_val.ctrl_grad.ct_direction
#define ctd_width	ct_val.ctrl_grad.ct_width
#define ctd_height	ct_val.ctrl_grad.ct_height
#define ctd_zoomflag	ct_val.ctrl_grad.ct_zoomflag
#define ctd_xzoom	ct_val.ctrl_grad.ct_xzoom
#define ctd_yzoom	ct_val.ctrl_grad.ct_yzoom
#define ctd_depth	ct_val.ctrl_grad.ct_depth
#define ctd_mode	ct_val.ctrl_grad.ct_mode

#define	ctb_color	ct_val.ctrl_bar.ct_color
#define	ctb_width	ct_val.ctrl_bar.ct_width
#define	ctb_start	ct_val.ctrl_bar.ct_start
#define	ctb_length	ct_val.ctrl_bar.ct_length


struct ctl_words {
        u_char ctl_type;
        char *ctl_string;
        int ctl_strlen;
};

extern const struct ctl_words ctl_words[];

struct page_attribute {
	u_int pg_flag;
	u_char pg_bimage[BUFSIZ]; /* background image */
	u_int pg_linenum;
	u_int pg_b_numcolor;	/* background gradation number of colors */
	u_int pg_b_dir;	 	/* background gradation deg */
} page_attribute[MAXPAGE];

#ifdef VFLIB
struct vfont {
	struct vfont *next;
	struct vfont *prev;
	struct vfont *lrunext;
	struct vfont *lruprev;
	u_int width;
	u_int height;
	u_int code;
	u_int minx;
	u_int charlen;
	char *fontname;
	XImage *image;
#ifdef ANTIALIAS
	u_char *dbitmap;
#endif
};
#endif /* VFLIB */

/*
 * The following variables are defined in global.c, and therefore
 * they are available in both "mgp" and "mgp2ps" binary.
 */
extern u_char *page_data[MAXPAGE][MAXLINE];
extern struct ctrl *page_control[MAXPAGE][MAXLINE];
extern struct ctrl *default_control[MAXLINE];
extern struct ctrl *tab_control[MAXTAB];
extern struct ctrl *init_control[MAXLINE];

extern u_int mgp_flag;
extern u_int maxpage;
extern u_int cur_page;
extern char *mgp_fname;

extern u_int parse_error;

extern Display *display;
extern Window window;
extern int screen;
extern u_int window_width;
extern u_int window_height;
extern Pixmap pixmap;

extern int char_size;
extern u_int horiz_gap;
extern u_int vert_gap;
extern u_int depth;
extern u_long fore_color;
extern u_long back_color;
extern u_long ctrl_color;

extern time_t t_start, t_fin;
extern u_int tbar_flag;

#define	VERT_GAP(s)		((s) * vert_gap / 100)
#define	HORIZ_GAP(s)		((s) * horiz_gap / 100)
#define	VERT_STEP(s)		((s) + VERT_GAP(s))
#define	HORIZ_STEP(s, x)	((x) + HORIZ_GAP(s))
#define	VERT_POS(y, s)		((y) + VERT_GAP(s)/2)


/*
 * The following variable are defined in x11.c.  Therefore, these are
 * accessible only in "mgp" binary, not in "mgp2ps".
 * We should separate header files.
 * (or, if mgp.c and print.c get merged the problem will vanish)
 */
extern GC gcall;	/* to copy; all plane */
extern GC gcnone;	/* to clear */
extern GC gcfore;	/* to make or data; foreground only */
extern GC gcback;	/* to clear text;  background only */
extern GC gcfb;		/* foreground and background color */
extern GC gcmask;	/* to mask; andInverted */
extern GC gcor;		/* or; use after mask */
extern GC gcpen;	/* pen */
extern GC gcgreen;	/* green */
extern GC gcyellow;	/* yellow */
extern GC gcred;	/* red */
extern GC gc_pl;	/* page list */
extern GC gc_plrev;	/* page list */
extern GC gc_pta;	/* page status ascii */
extern GC gc_ptk;	/* page status kanji */

/*
 * The following variable are defined in mgp.c.  Therefore, these are
 * accessible only in "mgp" binary, not in "mgp2ps".
 * We should separate header files.
 * (or, if mgp.c and print.c get merged the problem will vanish)
 */
extern Window plwin[MAXPAGE];
extern Pixmap maskpix;
extern XFontStruct *plfs;
extern XFontStruct *plkfs;
extern XFontStruct *plsfs;
extern u_int pause_flag;
extern u_int pg_mode;
extern u_int pl_mode;
extern u_int pl_tmode;
extern u_int pl_nx, pl_ny, pl_y;
extern u_long pl_fh, pl_fw;
extern u_long depth_mask;
extern u_char back_clname[MAXPATHLEN];
#ifdef	VFLIB
extern char *vfcap_name;
#endif

/*draw.c*/
extern void draw_page __P((u_int));

/*parse.c*/
extern void load_file __P((u_char *));
extern int ctlcmp __P((struct ctrl *, struct ctrl *));

/*plist.c*/
extern void pl_on __P((void));
extern void pl_off __P((void));
extern void pl_pdraw __P((int, GC));
extern void pl_title __P((int));
extern char *page_title __P((int));
extern void pg_on __P((u_int, u_int));
extern void pg_off __P((void));
extern void pl_config __P((void));
extern int draw_kstring __P((Drawable, char *, int, int));

/*font.c*/
extern struct vfont *vfc_get __P((u_int, u_int, u_int, int));
extern void vfc_setfont __P((char *));
#ifdef ANTIALIAS
extern XImage *vfc_image __P((struct vfont *, long, long));
#endif

/*x11.c or x11dummy.c*/
extern u_int window_x;
extern u_int window_y;
extern void init_win1 __P((void));
extern void init_win2 __P((void));
extern int get_color __P((u_char *, u_long *));

/* background.c */
extern void g_rotate __P((byte *, struct ctrl_grad *));
extern double cdist __P((int, int, int, int, int, int));
extern double lcdist __P((int, int, int, int, int, int, double, double));
extern byte *draw_gradation __P((struct ctrl_grad *));
extern Image *make_XImage __P((byte *, unsigned int, unsigned int));
extern struct g_color *name2gcolor __P((u_char *));
